/*
S.co.tt's Motion Control Slider
For more information:  http://s.co.tt/mc-slider
*/


#include <Wire.h>
#include <Math.h>
#include <LCD.h>
#include <LiquidCrystal_I2C.h>
#include <AccelStepper.h>

#define I2C_ADDR  0x27  
#define BACKLIGHT_PIN  3
#define En_pin  2
#define Rw_pin  1
#define Rs_pin  0
#define D4_pin  4
#define D5_pin  5
#define D6_pin  6
#define D7_pin  7

#define PUL_pin  13
#define DIR_pin  12

#define btnREV_pin   0
#define ledREV_pin   1
#define btnSTOP_pin  2
#define ledSTOP_pin  3
#define btnPLAY_pin  4
#define ledPLAY_pin  5
#define btnREC_pin   6
#define ledREC_pin   7
#define btnBNCE_pin  8
#define ledBNCE_pin  9
#define btnFWD_pin  10
#define ledFWD_pin  11

#define potSPD_apin  3

AccelStepper stepper(AccelStepper::DRIVER, PUL_pin, DIR_pin);
LiquidCrystal_I2C lcd(I2C_ADDR,En_pin,Rw_pin,Rs_pin,D4_pin,D5_pin,D6_pin,D7_pin);

int btnLT_apin = 1;
int btnRT_apin = 0;

long lStepsPerWhole = 0L;
long lStepsEdgeMargin = 0L;
float fTrackLengthInches = 30.75F;
long lTimePerWhole = 0;
float fStepsPerInch = 0.0F;
int iLastPotValue = 0;

void setup()
{
  setupLCD();
  displayIntro();
 
  delay(2000);

  displayLoading(); 
  setupStepper(); 
  setupInterface();
  
  // Skip calibration if STOP and REC are being held down;  This is mostly just used for testing.
  if (digitalRead(btnSTOP_pin) == HIGH && digitalRead(btnREC_pin) == HIGH && lStepsPerWhole == 0)
  {
    boolean bOK = calibrate();
    while (!bOK)
    {
      displayWaitCalibrationFailed();
      bOK = calibrate();
    }
    
    delay(1000);
  }
  //testVariousSpeeds();
  initLoopDisplay();
}

void setupLCD()
{
  lcd.begin(20, 4);
  lcd.setBacklightPin(BACKLIGHT_PIN, POSITIVE);
  lcd.setBacklight(HIGH);
}

void displayIntro()
{
  lcd.home ();
  lcd.print("S.CO.TT's");  
  lcd.setCursor(0, 1);
  lcd.print("MotionControlSlider");
  lcd.setCursor(0, 2);
  lcd.print("http://s.co.tt");
  lcd.setCursor(0, 3);        // go to the fourth line
  lcd.print("Version 0.1"); 
}

void displayLoading()
{
  lcd.setCursor(0, 2);
  lcd.print("");
  lcd.setCursor(0, 3);        // go to the fourth line
  lcd.print("Loading...     "); 
}

void setupStepper()
{
  digitalWrite(13, LOW);
  stepper.setSpeed(6000);
  stepper.setAcceleration(5000);
  stepper.setMaxSpeed(6000);
  stepper.setMinPulseWidth(20);  
}

void setupInterface()
{
  //Enable analog pullup resistors
  digitalWrite(A0, HIGH);
  digitalWrite(A1, HIGH);
  
  pinMode(btnREV_pin, INPUT);
  digitalWrite(btnREV_pin, HIGH);
  pinMode(btnSTOP_pin, INPUT);
  digitalWrite(btnSTOP_pin, HIGH);
  pinMode(btnPLAY_pin, INPUT);
  digitalWrite(btnPLAY_pin, HIGH);
  pinMode(btnREC_pin, INPUT);
  digitalWrite(btnREC_pin, HIGH);
  pinMode(btnBNCE_pin, INPUT);
  digitalWrite(btnBNCE_pin, HIGH);
  pinMode(btnFWD_pin, INPUT);
  digitalWrite(btnFWD_pin, HIGH);
  
  pinMode(ledREV_pin, OUTPUT);
  digitalWrite(ledREV_pin, HIGH);
  pinMode(ledSTOP_pin, OUTPUT);
  digitalWrite(ledSTOP_pin, LOW);
  pinMode(ledPLAY_pin, OUTPUT);
  digitalWrite(ledPLAY_pin, LOW);  
  pinMode(ledREC_pin, OUTPUT);
  digitalWrite(ledREC_pin, LOW);
  pinMode(ledBNCE_pin, OUTPUT);
  digitalWrite(ledBNCE_pin, LOW);
  pinMode(ledFWD_pin, OUTPUT);
  digitalWrite(ledFWD_pin, LOW);
  
  delay(100); 
  digitalWrite(ledSTOP_pin, HIGH);
  delay(100); 
  digitalWrite(ledREV_pin, LOW);
  delay(100); 
  digitalWrite(ledPLAY_pin, HIGH);
  delay(100); 
  digitalWrite(ledSTOP_pin, LOW);
  delay(100); 
  digitalWrite(ledREC_pin, HIGH);
  delay(100); 
  digitalWrite(ledPLAY_pin, LOW);
  delay(100); 
  digitalWrite(ledBNCE_pin, HIGH);
  delay(100); 
  digitalWrite(ledREC_pin, LOW);
  delay(100); 
  digitalWrite(ledFWD_pin, HIGH);
  delay(100); 
  digitalWrite(ledBNCE_pin, LOW);
  delay(100); 
  digitalWrite(ledFWD_pin, LOW);
}

boolean readAnalogButton(int pin)
{
  // If analog value < 640, assume button is pressed.
  return (analogRead(pin) < 640);
}

unsigned long moveToStopper(int dir)
{
  return moveToStopper(dir, false); 
}

unsigned long testVariousSpeeds()
{
  unsigned long lTemp = moveToStopper(1, true);
  
   lcd.clear();
    lcd.home();
  
  long lSpeed = 500;
  
  for (int i=1; i<=8; i++)
  {
    if (i == 3)
      lcd.setCursor(0, 1);
    if (i == 5)
      lcd.setCursor(0, 2);
    if (i == 7)
      lcd.setCursor(0, 3);
      
    stepper.setAcceleration(160000);

    lSpeed = lSpeed + 500;
    
    stepper.setMaxSpeed(lSpeed);
 
    lcd.print(lSpeed);
    lcd.print(":");
    
    long iMS = millis();
    
    //lcd.print(">>");
    //lcd.print(stepper.currentPosition());
    //lcd.print("<<");
    
    if (stepper.currentPosition() > 5000)
      stepper.runToNewPosition(lStepsEdgeMargin);
    else
      stepper.runToNewPosition(lStepsPerWhole - lStepsEdgeMargin);
    
    unsigned long iDuration = millis() - iMS;
    
    lcd.print(iDuration / 100);
    lcd.print(" ");
  }
  
  delay(300000);
}

unsigned long moveToStopper(int dir, boolean bCalibrate)
{
  unsigned long lStepsPerLoop = 1L;
  unsigned long lSteps = 0L;
  unsigned int iIter = 0;
  unsigned long startTime = millis();
  boolean bLight = true;
  
  int incre = 0;
  
  while(true)
  {
    stepper.move(lStepsPerLoop * dir);
    stepper.setSpeed(8000);
    //stepper.runSpeed();
    if (stepper.runSpeedToPosition())
      lSteps += lStepsPerLoop;
    
    if (dir < 0 && readAnalogButton(btnLT_apin))
    {
      digitalWrite(ledREV_pin, LOW);
      if (lSteps == lStepsPerLoop && bCalibrate)
        return 2;
      else 
        return lSteps;
    } 
    if (dir > 0 && readAnalogButton(btnRT_apin))
    {
      digitalWrite(ledFWD_pin, LOW);
      if (lSteps == lStepsPerLoop && bCalibrate)
        return 2;
      else
        return lSteps;
    }
    if (iIter == 750)
    {
      if (dir > 0)
        digitalWrite(ledFWD_pin, bLight);
      if (dir < 0)
        digitalWrite(ledREV_pin, bLight);
      bLight = !bLight;
      
      if (millis() - startTime > 60000L)
      {  
        digitalWrite(ledFWD_pin, LOW);
        digitalWrite(ledREV_pin, LOW);
  
        return 1;
      }  
      
      iIter = 0;
    }
    
    iIter++;
  }
  
  digitalWrite(ledFWD_pin, LOW);
  digitalWrite(ledREV_pin, LOW);
  
  return lSteps;
}

boolean calibrate()
{
  lcd.clear();
  lcd.home();
  lcd.print("Calibrating...");  
  lcd.home();
  /*
  if (readAnalogButton(btnLT_apin) || readAnalogButton(btnRT_apin))
  {
    lcd.setCursor(0, 1);
    lcd.print("Please move the");
    lcd.setCursor(0, 2);
    lcd.print("carriage away from");
    lcd.setCursor(0, 3);
    lcd.print("the edge..");
    
    
    boolean bDot = true;
    
    while(true)
    {
      if (!readAnalogButton(btnLT_apin) && !readAnalogButton(btnRT_apin))
      {
        break;
      } else {
        lcd.setCursor(10, 3);
        if (bDot == true) {
          lcd.print(".");
          bDot = false;
        } else {
          lcd.print("_");
          bDot = true;
        }
      }
      delay(250);
    }
  }
  */
  
  lcd.clear();
  lcd.home();
  lcd.print("Calibrating...");
  lcd.setCursor(0, 1);
  lcd.print("Moving end...");
  unsigned long lTemp = moveToStopper(1, true);
  if (lTemp > 1)
    lcd.print("OK");
  else
    lcd.print("FAIL");
  lcd.setCursor(0, 2);
  lcd.print("Moving reverse...");

  long lTime0 = millis();
  unsigned long lStepsRight = moveToStopper(-1);
  lTime0 = millis() - lTime0;
  if (lStepsRight > 1)
    lcd.print("OK");
  else
    lcd.print("FAIL");
  lcd.setCursor(0, 3);
  lcd.print("Moving forward...");
  
  long lTime1 = millis();
  unsigned long lStepsLeft = moveToStopper(1);
  lTime1 = millis() - lTime1;
  if (lStepsLeft > 1)
  {
    stepper.setCurrentPosition(0);
    lcd.print("OK");
  } else
    lcd.print("FAIL");
    
  lTimePerWhole = (lTime0 + lTime1) / 2;
  
  delay(250);
  
  lcd.clear();
  lcd.home();
  
  if (lStepsRight == 0 || lStepsLeft == 0)
  {
//             |------------------|
    lcd.print("Calibration Failed!");
    lcd.setCursor(0, 1);
    lcd.print("Error: No steps");
    lcd.setCursor(0, 2);
    lcd.print("found when moving:");
    lcd.setCursor(0, 3);
    if (lStepsRight == 0)
      lcd.print("RIGHT ");
    if (lStepsLeft == 0)
      lcd.print("LEFT");
      
    return false;
  }
  
  if (lStepsRight == 1 || lStepsLeft == 1)  // Using the # one as an error condition because there's no way 1 step can take it to the end of the carriage.
  {
//             |------------------|
    lcd.print("Calibration Failed!");
    lcd.setCursor(0, 1);
    lcd.print("Error: Timeout");
    lcd.setCursor(0, 2);
    lcd.print("when moving:");
    lcd.setCursor(0, 3);
    if (lStepsRight == 1)
      lcd.print("RIGHT ");
    if (lStepsLeft == 1)
      lcd.print("LEFT");
      
    return false;
  }
  
  long lDiff = lStepsRight - lStepsLeft;
  lDiff = abs(lDiff);  // Arduino docs say not to do math in func call
  
  lStepsPerWhole = (lStepsRight + lStepsLeft) / 2;
  float fDiffPct = ((float)lDiff / (float)lStepsPerWhole) * 100;

  if (fDiffPct > 1)
  {
//             |------------------|
    lcd.print("Error: Steps diff.");
    lcd.setCursor(0, 1);
    lcd.print("L:");
    lcd.print(lStepsLeft);
    lcd.print(" diff:");
    lcd.print(lDiff);
    lcd.setCursor(0, 2);
    lcd.print("R:");
    lcd.print(lStepsRight);
    lcd.print("  avg:");
    lcd.print(lStepsPerWhole);
    lcd.setCursor(0, 3);
    lcd.print("PCT:");
    lcd.print(fDiffPct);
    return false;    
  }
  
  fStepsPerInch = (float)lStepsPerWhole / fTrackLengthInches;
  
  lcd.print("Calibration OK!");
  lcd.setCursor(0, 1);
  lcd.print("Steps: ");
  lcd.print(lStepsPerWhole);
  lcd.setCursor(0, 2);
  lcd.print("Len: ");
  lcd.print(fTrackLengthInches);
  lcd.setCursor(10, 2);
  lcd.print("in.");
  lcd.setCursor(0, 3);
  lcd.print("s/in: ");
  lcd.print(fStepsPerInch);
  
  lStepsEdgeMargin = lStepsPerWhole * 0.025;
  
  stepper.setCurrentPosition(lStepsPerWhole);
  
  stepper.runToNewPosition(lStepsPerWhole - lStepsEdgeMargin);
  
  return true;
}

void displayWaitCalibrationFailed()
{
  delay(5000);
  lcd.home();
  lcd.print("Bounce to recal...");
  
  int incre = 0;
  while (true)
  {
    if (incre == 20)
      digitalWrite(ledBNCE_pin, HIGH);
      
    if (incre == 40)
    {
      digitalWrite(ledBNCE_pin, LOW);
      incre = 0;
    }
    
    int val = digitalRead(btnBNCE_pin);
    if (val == LOW)
    {
      digitalWrite(ledBNCE_pin, LOW);
      break;
    }
    delay(50);
    incre++;
  }  
}

// Mode: 1=REV, 2=STOP, 3=PLAY, 5=BOUNCE, 6=FWD
int iMode = 2;
int iModePrev = -1;
int iModeRec = 0;
boolean bRecording = false;
boolean bPlaying = false;
boolean bBounce = false;
int iter = 0;
int iSpeed = 0;
int iSpeedPrev = 0;
int iSpeedRec = 0;

int iStepsForAccel = 0;

unsigned long lRecorder[32][3];
int iRecorderIndex = -1;
unsigned long lRecStartMS = 0;
unsigned long lRecSpeedMS = 0;
int iPlayerIndex = 0;
unsigned long lPlayStartMS = 0;

void loop()
{
  readLoopButtons(iter);
  
  if (bPlaying)
  {
    if (iPlayerIndex == 0 || millis() >= lPlayStartMS + lRecorder[iPlayerIndex][0])
    {

      iMode = lRecorder[iPlayerIndex][1];
      iSpeed = lRecorder[iPlayerIndex][2];
      
      if (iSpeed != iSpeedPrev)
        setSpeed(iSpeed);

      iPlayerIndex++;
      
      if (iPlayerIndex > iRecorderIndex + 1)
      {
        iMode = 2;
        lcd.setCursor(0, 3);
        lcd.print("PLAYBACK DONE.");
        digitalWrite(ledPLAY_pin, LOW);
        bPlaying = false;
        readLoopButtons(8);
      }
    }
  }
  
  if (iMode != iModePrev)
  {
    clearLoopButtons();  
    
    if (iModePrev == 2)
    {
      //LCD Updates are very costly; Only do them when coming out of stop.
      initLoopDisplay();
    }

    if (iMode == 2)
    {
      bBounce = false;
      digitalWrite(ledBNCE_pin, LOW);
      digitalWrite(ledSTOP_pin, HIGH);
      stepper.stop();
      stepper.runToPosition();
      initLoopDisplay();
    } else {
      if (iMode == 6)
      {
        stepper.moveTo(lStepsPerWhole - lStepsEdgeMargin);
        unsigned int lStepperAccelTo = stepper.currentPosition() + iStepsForAccel;
        digitalWrite(ledFWD_pin, HIGH);
        if (!bRecording && !bPlaying)
        while (stepper.currentPosition() <= lStepperAccelTo && stepper.currentPosition() < stepper.targetPosition())  // Blocking run operation during accel for smoothness.
          stepper.run();
      } 
      if (iMode == 1)
      {
        stepper.moveTo(lStepsEdgeMargin);
        unsigned int lStepperAccelTo = stepper.currentPosition() - iStepsForAccel;
        digitalWrite(ledREV_pin, HIGH);
        if (!bRecording && !bPlaying)
        while (stepper.currentPosition() >= lStepperAccelTo && stepper.currentPosition() > stepper.targetPosition())  // Blocking run operation during accel for smoothness.
          stepper.run();
      }
    } 
  }
  
  stepper.run();
  
  iModePrev = iMode;  
  
  if (stepper.currentPosition() == stepper.targetPosition())
  {
    if (bBounce && iMode == 1)
      iMode = 6;
    else if (bBounce && iMode == 6)
      iMode = 1;
    else
      iMode = 2;
  }
  
  iter++;
  
  if (iter > 8)
  {
    if (bRecording)
    {
      if (iMode != iModeRec)
        recordState();
        
      if (iSpeed != iSpeedRec && millis() - lRecSpeedMS > 333)
      {
        lRecSpeedMS = millis();
        recordState(); 
      }
    }
    iter = 0;
  }
}

void recordState()
{
  iRecorderIndex++;
  
  unsigned long lTime = millis() - lRecStartMS;
  
  iModeRec = iMode;
  iSpeedRec = iSpeed;
  
  if (iRecorderIndex == 0)
    lRecorder[0][0] = stepper.currentPosition();
  else
    lRecorder[iRecorderIndex][0] = lTime;
  lRecorder[iRecorderIndex][1] = iMode;
  lRecorder[iRecorderIndex][2] = iSpeed;
  
  if (iRecorderIndex > 30)
  {
    iMode = 2;  
    bRecording = false;
    digitalWrite(ledREC_pin, LOW);
  }
  
  /* For debugging:
  lcd.setCursor(0, 1);
  lcd.print("T:");
  lcd.print(lTime);
  lcd.print(" M:");
  lcd.print(iMode);
  lcd.print(" S:");
  lcd.print(iSpeed);
  */
}

void playHome()
{
  lcd.setCursor(0, 3);
  lcd.print("Going to home pos..");
  setSpeed(1023);
  stepper.runToNewPosition(lRecorder[0][0]);
  lcd.setCursor(0, 3);
  lcd.print("PLAYING             ");
}

void initLoopDisplay()
{
  lcd.clear();
  lcd.home();
  
  switch (iMode)
  {
    case 1: //REV
      lcd.print("Moving reverse..");
      break;
    case 2: //STOP
      lcd.print("Stopped.");
      break;
    case 6: //FWD
      lcd.print("Moving forward..");
      break;  
  }  
  if (bPlaying)
  {
    lcd.setCursor(0, 3);
    lcd.print("PLAYING          "); 
  }
  if (bRecording)
  {
    lcd.setCursor(0, 3);
    if (iRecorderIndex > 30)
      lcd.print("RECORDER FULL");
    else
      lcd.print("RECORDING"); 
  }
  if (bBounce)
  {
    lcd.home();
    lcd.print("Bouncing..         ");  
  }
}

void refreshLoopDisplay()
{
  lcd.setCursor(0, 2);
  lcd.print("iSpd: ");
  lcd.print(iSpeed);
  lcd.setCursor(0, 3);
  lcd.print("pos:");
  lcd.print(stepper.currentPosition());
  lcd.print(" spd:");
  lcd.print(stepper.speed());
}

void readLoopButtons(int iter)
{
  switch (iter)
  {
    case 0:
      if (digitalRead(btnREV_pin) == LOW)
      {
        iMode = 1;
      }
      break; 
    case 1:
      if (digitalRead(btnSTOP_pin) == LOW)
      {
        iMode = 2;
        bPlaying = false;
        digitalWrite(ledPLAY_pin, LOW);
      } 
      break; 
    case 2:
      if (digitalRead(btnPLAY_pin) == LOW)
      {
        if (iRecorderIndex > -1)
        {
          playHome();
          bPlaying = true;
          bRecording = false;
          lPlayStartMS = millis();
          iPlayerIndex = 0;
          digitalWrite(ledPLAY_pin, HIGH);
          digitalWrite(ledREC_pin, LOW);
        } else {
          lcd.setCursor(0, 1);
          lcd.print("Nothing recorded.");
        }
      }  
      break; 
    case 3:
      if (digitalRead(btnREC_pin) == LOW)
      {
        if (millis() - lRecStartMS > 500)
        {
          if (bRecording)
          {
            bRecording = false;
            lRecStartMS = millis();
            initLoopDisplay();
          } else {
            bRecording = true;
            bPlaying = false;
            lRecStartMS = millis();
            iRecorderIndex = -1;
            digitalWrite(ledPLAY_pin, LOW);
          }
          digitalWrite(ledREC_pin, bRecording);
        }
      } 
      break; 
    case 4:
      if (digitalRead(btnBNCE_pin) == LOW)
      {
        if (bBounce)
          bBounce = false;
        else {
          bBounce = true;
          iMode = 1;
          digitalWrite(ledREV_pin, HIGH);
        }
        digitalWrite(ledBNCE_pin, bBounce);
      } 
      break; 
    case 5:
      if (digitalRead(btnFWD_pin) == LOW)
      {
        iMode = 6;
      }  
      break; 
    case 6:
      if (iMode == 6 && readAnalogButton(btnRT_apin))
      {
        iMode = 2; 
        stepper.stop();
        stepper.runSpeedToPosition();
        stepper.setCurrentPosition(lStepsPerWhole);
        stepper.moveTo(lStepsPerWhole - lStepsEdgeMargin);
        stepper.runToPosition();
      }
      break; 
    case 7:
      if (iMode == 1 && readAnalogButton(btnLT_apin))
      {
        iMode = 2;
        stepper.stop();
        stepper.runSpeedToPosition();
        stepper.setCurrentPosition(0);
        stepper.moveTo(lStepsEdgeMargin);
        stepper.runToPosition();
      }
      break; 
    case 8:
      if (!bPlaying)
      {
        int iSpeedNew = analogRead(potSPD_apin);
  
        if (iSpeedNew > iSpeedPrev + 3 || iSpeedNew < iSpeedPrev - 3)  //Prevent jitter -- resolution of 6.
        {
          setSpeed(iSpeedNew);
        }
      }
      break; 
  }
  
  return;
}

void setSpeed(int iSpeedNew)
{
  float fAccelMult = 1.8;
  float fAccelTime = (1.0 / fAccelMult);
  float fMaxSpeed = 3.2 * (float)iSpeedNew;
  float fAccel = fAccelMult * fMaxSpeed;
  stepper.setAcceleration(fAccel);
  stepper.setMaxSpeed(fMaxSpeed);
  
  iStepsForAccel = (int)(0.5 * fAccel * (fAccelTime * fAccelTime));
  
  iSpeedPrev = iSpeed;
  iSpeed = iSpeedNew; 
  
  if (iMode == 2)
  {
    // lTimePerWhole, fStepsPerInch
    float fInchesPerSec = round(fMaxSpeed / fStepsPerInch * 10) / 10;
    lcd.setCursor(0, 2);
    lcd.print("in/sec: ");
    lcd.print(fInchesPerSec);
  }
}

void clearLoopButtons()
{
  switch (iModePrev)
  {
    case 1:
      digitalWrite(ledREV_pin, LOW);
      break;
    case 2:
      digitalWrite(ledSTOP_pin, LOW);
      break;
    case 6:
      digitalWrite(ledFWD_pin, LOW);
      break;  
  }
}
